#!/usr/bin/env python3

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import PyInstaller.__main__

src = "blackboard_sync"
is_win = sys.platform.startswith('win')
sep = ';' if is_win else ':'

icon = f'{src}/assets/logo'
icon += '.ico' if is_win else '.png'

assets = f'{src}/assets'
assets = assets + sep + assets

qt = f'{src}/qt'
qt = qt + sep + qt

uni = f'{src}/universities.json'
uni += sep + src

PyInstaller.__main__.run([
    f'--icon={icon}',
    f'--add-data={assets}',
    f'--add-data={qt}',
    f'--add-data={uni}',
    '--noconfirm',
    '--clean',
    '--windowed',
    '--name=BBSync',
    'main.py',
])



"""
Parse an html to look for links
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from typing import List, NamedTuple
from bs4 import BeautifulSoup

class Link(NamedTuple):
    href: str
    text: str

class ContentParser:
    def __init__(self, body: str, base_url: str):
        links = []
        soup = BeautifulSoup(body, 'html.parser')

        for link in soup.find_all('a'):
            # Add link for later download
            links.append(Link(href=link.get('href'), text=link.text.strip()))

            # Replace for local instance
            if link['href'].startswith(base_url):
                filename = link.text.strip()
                link['href'] = filename
                link.string = filename
        self._links = links
        self.soup = soup

    @property
    def links(self) -> List[Link]:
        return self._links

    @property
    def body(self) -> str:
        return str(self.soup)

def validate_webdav_response(response, link: str, base_url: str):
    if response.status_code == 200:
        h = response.headers
        content_type = h.get('Content-Type', '')
        content_len = int(h.get('Content-Length', 0))

        # TODO: feature: select mime types
        len_limit = 1024 * 1024 * 20 # 20 MB
        return link.startswith(base_url) and 'video' not in content_type and content_len < len_limit
    return False



"""
Blackboard API.

an interface to make Blackboard REST API calls on a session basis
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License v2
# as published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License v2
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.

import logging
from typing import Union, Optional
from functools import wraps
from collections.abc import Callable

import requests
from requests.cookies import RequestsCookieJar

from .blackboard import (BBCourse, BBAttachment, BBMembership, BBContentChild,
                         BBCourseContent)


class SafeFormat(dict):
    """Custom dictionary object.

    Needed to safely format endpoint strings with placeholders
    without having all parameters (those not present will be left blank)
    """

    def __missing__(self, key):
        """Return empty string for missing entries."""
        return ''


class BlackboardSession:
    """Represents a user session in Blackboard."""

    _logger = logging.getLogger(__name__)
    _logger.setLevel(logging.DEBUG)
    _logger.addHandler(logging.NullHandler())

    def __init__(self, base_url: str, cookies: RequestsCookieJar):
        # Root API URL
        self._base_url = base_url
        # Use cookies for requests
        self._bb_session = requests.Session()
        self._bb_session.cookies = cookies
        # Obtain username by calling self
        self._username : Optional[str] = None
        self._timeout = 12

    def get(endpoint: str, version: int = 1, use_api: bool = True, json: bool = True, **g_kwargs):
        """Return a decorator (needed to use fancy notation).

        :param string endpoint: Endpoint to make API call to, including placeholders
        :param int version: Version of the BB API used at endpoint (used as part of url)
        :param bool json: If false, returns raw requests response, otherwise returns JSON Object
        :param dict g_kwargs: Any argument in this parameter will be passed on to the requests call
        """

        def get_decorator(func: Callable) -> Callable:
            """Return wrapped function.

            :param function func: Function to decorate
            """

            @wraps(func)
            def get_wrapper(self, *args, **kwargs):
                """Wrap function in REST API call.

                :param any id: Not in current use
                :param list args: Not in current use
                :param dict kwargs: Any contents will be passed as request parameters
                """
                param_endpoint = endpoint.format_map(SafeFormat(kwargs))

                api_root = f"/learn/api/public/v{version}"
                endpoint_format = f"{self._base_url}{api_root}{param_endpoint}"

                if not use_api:
                    endpoint_format = param_endpoint
                if endpoint_format[-1] == "/":
                    endpoint_format = endpoint_format[:-1]

                self.logger.debug(f"Making request to {endpoint_format}")
                response = self._bb_session.get(endpoint_format, params=kwargs,
                                                timeout=self._timeout, **g_kwargs)
                if json:
                    response = response.json()
                else:
                    return func(self, response)

                if not response:
                    raise ValueError('Server response empty')
                elif 'status' in response:
                    self.logger.error(f"Code {response['status']}: {response['message']}")
                    if response['status'] == 401:
                        raise ValueError('Not authorized')
                    if response['status'] == 403:
                        if 'code' in response and response['code'] == 'bb-rest-course-is-private':
                            raise ValueError('Private course')
                    raise ValueError('Server responded with an error code')
                elif 'results' in response:
                    return func(self, response['results'])
                else:
                    return func(self, response)
            return get_wrapper
        return get_decorator

    @property
    def username(self) -> str:
        """Username field used for API requests."""
        if self._username is None:
            self._username = self.fetch_users(user_id='me')['id']
        return self._username

    @property
    def timeout(self) -> int:
        """Request timeout."""
        return self._timeout

    @timeout.setter
    def timeout(self, t: int) -> None:
        self._timeout = t

    # WEBDAV DOWNLOAD

    @get("{webdav_url}", stream=True, json=False, use_api=False)
    def download_webdav(self, response):
        """Downloads an arbitrary webdav file"""
        return response


    # API CALLS
    # https://developer.blackboard.com/portal/displayApi

    # announcements #

    # Get Announcement(s)
    @get("/announcements/{announcement_id}")
    def fetch_announcements(self, response):
        """Return a list of System Announcements.

        :param announcement_id: Get a System Announcement.
        """
        return response

    # attendance #

    # Get Course Meetings
    @get("/courses/{course_id}/meetings")
    def fetch_course_meetings(self, response):
        """Return a list of course meetings for a given course id.

        :param course_id: The course or organization ID.
        """
        return response

    # Generate Attendance Data Download Url
    @get("/courses/{course_id}/meetings/downloadUrl")
    def fetch_attendance_data_download_url(self, response):
        """Generate Download URL for Attendance Data.

        :param course_id: The course or organization ID.
        """
        return response

    # Get Attendance Records By User Id
    @get("/courses/{course_id}/meetings/users/{user_id}")
    def fetch_attendance_records_by_user_id(self, response):
        """Return a list of Course Meeting Attendance for a given user id
        regardless of courses and meetings.

        :param course_id: The course or organization ID.
        :param user_id: The user ID.
        """
        return response

    # Get Course Meeting
    @get("/courses/{course_id}/meetings/{meeting_id}")
    def fetch_course_meeting(self, response):
        """Return a Course Meeting for the given meeting Id.

        :param course_id: The course or organization ID.
        :param meeting_id:
        """
        return response

    # Get Attendance Records By Meeting Id / Get Attendance Record
    @get("/courses/{course_id}/meetings/{meeting_id}/users/{user_id}")
    def fetch_attendance_records_by_meeting_id(self, response):
        """Return a Course Meeting Attendance information for the given meeting and user Id.

        :param course_id: The course or organization ID.
        :param meeting_id:
        :param user_id: The user ID.
        """
        return response

    # calendar #

    # Get Calendars
    @get("/calendars/")
    def fetch_calendar(self, response):
        """Get the list of calendars. This endpoint will return all calendars viewable by the user.

        All users can request a list of calendars viewable to them.
        """
        return response

    # Get Calendar Item(s)
    @get("/calendars/items/{calendar_item_type}/{calendar_item_id}")
    def fetch_calendar_items(self, response):
        """Get a course calendar item.

        :param calendar_item_type: One of (Course, GradebookColumn, Institution,
            OfficeHours, Personal).
        :param calendar_item_id:
        """
        return response

    # content #

    # Get Content(s)
    @get("/courses/{course_id}/contents/{content_id}")
    def fetch_contents(self, response) -> list[BBCourseContent]:
        """List top-level content items in a course.

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        """
        return [BBCourseContent(**content) for content in response]

    # Get Content Children
    @get("/courses/{course_id}/contents/{content_id}/children")
    def fetch_content_children(self, response) -> list[BBContentChild]:
        """List all child content items directly beneath another content item.

        This is only valid for content items that are allowed to have children (e.g. Folders).

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        """
        return [BBContentChild(**child) for child in response]

    # content file attachments #

    # Get File Attachment(s)
    @get("/courses/{course_id}/contents/{content_id}/attachments/{attachment_id}")
    def fetch_file_attachments(self, response) -> list[BBAttachment] | BBAttachment:
        """Get the file attachment meta data associated to the Content Item.

        or Get the file attachment meta data by an attachment ID.

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        :param attachment_id:
        """
        if isinstance(response, list):
            return [BBAttachment(**a) for a in response]
        else:
            return BBAttachment(**response)

    # Get File Attachment(s)
    @get("/courses/{course_id}/contents/{content_id}/attachments/{attachment_id}/download",
         json=False, stream=True)
    def download(self, response):
        """Download the contents of a Content Item.

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        :param attachment_id:
        """
        return response

    # content group assignments #

    # Get Content Group(s)
    @get("/courses/{course_id}/contents/{content_id}/groups/{group_id}")
    def fetch_content_groups(self, response):
        """Return a list of content group associations for the specified content.

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        :param group_id: The group ID.
        """
        return response

    # content resources #

    # Get Course Resource(s)
    @get("/courses/{course_id}/resources/{resource_id}")
    def fetch_course_resources(self, response):
        """Return a list of the top-level course resources.

        or Load a Course Resource by Id.

        :param course_id: The course or organization ID.
        :param resource_id: The xythos resource ID.
        """
        return response

    # Get Course Resource Children
    @get("/courses/{course_id}/resources/{resource_id}/children")
    def fetch_course_resource_children(self, response):
        """
        Return a list of Course Resources that are children of the specified Resource.

        :param course_id: The course or organization ID.
        :param resource_id: The xythos resource ID.
        """
        return response

    # content review #

    # Get Review Status(s)
    @get("/courses/{course_id}/contents/{content_id}/users/{user_id}/reviewStatus")
    def fetch_review_status(self, response):
        """Obtain the review status for a content item.

        This endpoint will only fetch the reviewStatus if the corresponding content
        was previously marked as reviewable.

        :param course_id: The course or organization ID.
        :param content_id: The Content ID.
        :param user_id: The user ID.
        """
        return response

    # course announcements #

    # Get Course Announcement(s)
    @get("/courses/{course_id}/announcements/{announcement_id}")
    def fetch_course_announcements(self, response):
        """Return a list of Course Announcements or Get a Course Announcement.

        :param course_id: The course or organization ID.
        :param announcement_id:
        """
        return response

    # course assessments #

    # Get Course Assessments Question(s)
    @get("/courses/{course_id}/assessments/{assessment_id}/questions/{question_id}")
    def fetch_questions(self, response):
        """Get the list of questions for an Ultra Assessment or Get a question by Id from it.

        :param course_id: The course or organization ID.
        :param assessment_id:
        :param question_id:
        """
        return response

    # course categories #

    # Get Categorie(s)
    @get("/catalog/categories/{category_type}/{category_id}")
    def fetch_category(self, response):
        """Return a list of categories of the provided type (course or organization).

        / Return the category corresponding the provided type (course or organization) and ID.

        :param category_type: One of (Course, Organization).
        :param category_id:
        """
        return response

    # Get Membership(s)
    @get("/catalog/categories/{category_type}/{category_id}/courses/{course_id}")
    def fetch_memberships(self, response):
        """
        Get courses associated with the provided category.

        :param category_type: One of (Course, Organization).
        :param category_id:
        """
        return response

    # Get Child Categories
    @get("/catalog/categories/{category_type}/{parent_id}/children")
    def fetch_child_categories(self, response):
        """Return a list of categories which are children of the category...

        corresponding to the provided type (course or organization) and Id

        :param category_type: One of (Course, Organization).
        :param parent_id:
        """
        return response

    # Get Memberships
    @get("/courses/{course_id}/categories")
    def fetch_categories(self, response):
        """Get categories associated with the provided course.

        :param course_id: The course or organization ID.
        """
        return response

    # course grade attempts #

    # Get Attempt File Meta Data (List)
    @get("/courses/{course_id}/gradebook/attempts/{attempt_id}/files/{attempt_file_id}")
    def fetch_attempt_file_metadata(self, response):
        """Get the list of file metadata for a Submission associated to the course and attempt.

        Get the file metadata for a Student Submission associated to the course
        and attempt.

        :param course_id: The course or organization ID.
        :param attempt_id:
        :param attempt_file_id:
        """
        return response

    # Get Attempt File Meta Data (List)
    @get("/courses/{course_id}/gradebook/attempts/{attempt_id}/files/{attempt_file_id}/download")
    def download_attempt_file_metadata(self, response):
        """Download the contents of the file for a Student Submission.

        :param course_id: The course or organization ID.
        :param attempt_id:
        :param attempt_file_id:
        """
        return response

    # course grade notations #

    # Get Grade Notation(s)
    @get("/courses/{course_id}/gradebook/gradeNotations/{grade_notation_id}")
    def fetch_grade_notations(self, response):
        """Return a list of grade notations. / Return a specific grade notation.

        :param course_id: The course or organization ID.
        :grade_notation_id:
        """
        return response

    # course gradebook categories #

    # Get Gradebook Categori(es)
    @get("/courses/{course_id}/gradebook/categories/{category_id}")
    def fetch_gradebook_categories(self, response):
        """Return a list of gradebook categories in a particular course.

        / Return the details of a gradebook category.

        :param course_id: The course or organization ID.
        :param category_id: the ID of the category to return
        """
        return response

    # course grades #

    # Get Grade Schema(s)
    @get("/courses/{course_id}/gradebook/schemas/{schema_id}")
    def fetch_grade_schemas(self, response):
        """Return a list of grade schemas associated with the specified course.

        / Load the grade schema associated with the specified course and schema Id.

        :param course_id: The course or organization ID.
        :param schema_id: The grade schema ID.
        """
        return response

    # Get Grade Column(s)
    @get("/courses/{course_id}/gradebook/columns/{column_id}", 2)
    def fetch_grade_columns(self, response):
        """Return a list of grade columns. / Load a specific grade column.

        :param course_id: The course or organization ID.
        :param column_id: The grade column ID.
        """
        return response

    # Get Grade Column Attempt(s)
    @get("/courses/{course_id}/gradebook/columns/{column_id}/attempts/{attempt_id}", 2)
    def fetch_column_attempts(self, response):
        """Return a list of attempts associated with the specified grade column.

        / Load the grade column attempt for the specified id.

        :param course_id: The course or organization ID.
        :param column_id: The grade column ID.
        :param attempt_id:
        """
        return response

    # Get Column Grade(s)
    @get("/courses/{course_id}/gradebook/columns/{column_id}/users/{user_id}", 2)
    def fetch_column_grades(self, response):
        """Return a list of grades associated with the specified grade column.

        / Load the grade column grade for a specific user.

        :param course_id: The course or organization ID.
        :param column_id: The grade column ID.
        :param user_id: The user ID.
        """
        return response

    # Get Column Grades Last Changed
    @get("/courses/{course_id}/gradebook/columns/{column_id}/users/lastChanged", 2)
    def fetch_column_grade_last_changed(self, response):
        """Load the grade column grade with the maximum change index.

        :param course_id: The course or organization ID.
        :param column_id: The grade column ID.
        """
        return response

    # Get User Grades
    @get("/courses/{course_id}/gradebook/users/{user_id}", 2)
    def fetch_user_grades(self, response):
        """Load the course grades for a specific user.

        :param course_id: The course or organization ID.
        :param user_id: The user ID.
        """
        return response

    # course grading periods #

    # Get Grading Period(s)
    @get("/courses/{course_id}/gradebook/periods/{period_id}")
    def fetch_grading_periods(self, response):
        """Return a list of grading periods. / Return a specific grading period.

        :param course_id: The course or organization ID.
        :param period_id:
        """
        return response

    # course group users #

    # Get Group Memberships
    @get("/courses/{course_id}/groups/{group_id}/users/{user_id}", 2)
    def fetch_group_memberships(self, response):
        """Return a list of group memberships objects for the specified group.

        / Loads a group membership in the specified group.

        :param course_id: The course or organization ID.
        :param group_id: The group ID.
        :param user_id: The user ID.
        """
        return response

    # course groups #

    # Get Group(s)
    @get("/courses/{course_id}/groups/{group_id}", 2)
    def fetch_groups(self, response):
        """Return a list of all top-level groups in the specified course.

        / Load a group in the specified course.

        :param course_id: The course or organization ID.
        :param group_id: The group ID.
        """
        return response

    # Get Group Set(s)
    @get("/courses/{course_id}/groups/sets/{group_id}", 2)
    def fetch_group_sets(self, response):
        """Return a list of all groupsets / Load a groupset in the specified course.

        :param course_id: The course or organization ID.
        :param group_id: The group ID.
        """
        return response

    # Get Group Set Children
    @get("/courses/{course_id}/groups/sets/{group_id}/groups", 2)
    def fetch_group_set_children(self, response):
        """Return a list of all groups within a groupset.

        :param course_id: The course or organization ID.
        :param group_id: The group ID.
        """
        return response

    # course memberships #

    # Get Course Membership(s)
    @get("/courses/{course_id}/users/{user_id}")
    def fetch_course_memberships(self, response):
        """Return a list of user memberships for the specified course or organization.

        / Load a user membership in the specified course.

        :param course_id: The course or organization ID.
        :param user_id: The user ID.
        """
        return response

    # Get User Memberships
    @get("/users/{user_id}/courses")
    def fetch_user_memberships(self, response) -> list[BBMembership]:
        """Return a list of course and organization memberships for the specified user.

        :param user_id: The user ID.
        """
        return [BBMembership(**memb) for memb in response]

    # courses #

    # Get Course Child(ren)
    @get("/courses/{course_id}/children/{child_course_id}")
    def fetch_course_children(self, response):
        """Return a list of course cross-listings.

        / Load a specific course cross-listing.

        :param course_id: The course or organization ID.
        :param child_course_id: The course or organization ID.
        """
        return response

    # Get Cross List Set
    @get("/courses/{course_id}/crossListSet")
    def fetch_cross_list_set(self, response):
        """Return the course cross-listing set for the specified course.

        This will return any and all parent/child associations regardless of the specified course
        being a parent or child course. The result will be empty if the specified course is
        not cross-listed.

        :param course_id: The course or organization ID.
        """
        return response

    # Get Task
    @get("/courses/{course_id}/tasks/{task_id}")
    def fetch_task(self, response):
        """Check the status of a queued task associated with a Course.

        Returns 200 unless task is complete.

        :param course_id: The course or organization ID.
        :param task_id:
        """
        return response

    # Get Course(s)
    @get("/courses/{course_id}", 3)
    def fetch_courses(self, response) -> Union[BBCourse, list[BBCourse]]:
        """Return a list of courses and organizations.

        / Loads a specific course or organization.

        :param course_id: The course or organization ID.
        """
        if isinstance(response, list):
            return [BBCourse(**course) for course in response]
        else:
            return BBCourse(**response)

    # data sources #

    # Get Data Source(s)
    @get("/dataSources/{data_source_id}")
    def fetch_data_sources(self, response):
        """Return a list of data sources.

        :param data_source_id: The data source ID.
        """
        return response

    # institutional hierarchy #

    # Get Associated Nodes
    @get("/courses/{course_id}/nodes")
    def fetch_associated_nodes(self, response):
        """Obtain a list of nodes to which a given course is directly associated.

        :param course_id: The course or organization ID.
        """
        return response

    # Get Node(s)
    @get("/institutionalHierarchy/nodes/{node_id}")
    def fetch_nodes(self, response):
        """Return the Top-level institutional hierarchy nodes.

        / Return the institutional hierarchy node corresponding the provided ID.

        :param node_id: The node ID.
        """
        return response

    # Get Node Children
    @get("/institutionalHierarchy/nodes/{node_id}/children")
    def fetch_node_children(self, response):
        """Return the children of the institutional hierarchy node corresponding...

        to the provided ID.

        :param node_id: The node ID.
        """
        return response

    # Get Node Course Associations
    @get("/institutionalHierarchy/nodes/{node_id}/courses")
    def fetch_node_course_associations(self, response):
        """Return a list of node-course relationships for the specified node.

        :param node_id: The node ID.
        """
        return response

    # lti #

    # Get Placement(s)
    @get("/lti/placements/{placement_id}")
    def fetch_placements(self, response):
        """Return a list of LTI placements.

        / Returns the LTI placement with the specified Id.

        :param placement_id:
        """
        return response

    # Get Domain Config(s)
    @get("/lti/domains/{domain_id}")
    def fetch_domain_config(self, response):
        """Return the list of LTI domain configs.

        / This endpoint returns the LTI domain config with the specified Id.

        :param domain_id:
        """
        return response

    # performance dashboard #

    # Get Review Status By Course Id
    @get("/courses/{course_id}/performance/contentReviewStatus")
    def fetch_performance_review_status(self, response):
        """List the content review statuses for all the users enrolled in a course.

        :param course_id: The course or organization ID.
        """
        return response

    # proctoring #

    # Get Proctoring Service(s)
    @get("/proctoring/services/{service_id}")
    def fetch_proctoring_services(self, response):
        """Return a list of proctoring service.

        / Return the proctoring service with the specified Id.

        :param service_id:
        """
        return response

    # roles #

    # Get Course Role(s)
    @get("/courseRoles/{role_id}")
    def fetch_course_roles(self, response):
        """Return a list of course roles. / Return a single course role.

        :param role_id: The course role ID.
        """
        return response

    # Get Institution Role(s)
    @get("/institutionRoles/{role_id}")
    def fetch_institution_roles(self, response):
        """Return a list of institution roles. / Load a specific institution role.

        :param role_id: The institution role ID.
        """
        return response

    # Get System Role(s)
    @get("/systemRoles/{role_id}")
    def fetch_system_roles(self, response):
        """Return a list of system roles. / Get a specific system role by roleId.

        :param role_id: The System Role ID.
        """
        return response

    # sessions #

    # Get Active Sessions
    @get("/sessions")
    def fetch_sessions(self, response):
        """List active user sessions in Learn."""
        return response

    # SIS logs #

    # Get SIS Logs By Data Set Uid
    @get("/logs/sis/dataSets/{id}")
    def fetch_sis_logs(self, response):
        """Return a list of SIS Integration logs.

        :param id: dataSetUid of the integration
        """
        return response

    # system #

    # Get Policies
    @get("/system/policies/privacy")
    def fetch_policies(self, response):
        """Return the links to the Blackboard and Institution privacy policies."""
        return response

    # Get System Task
    @get("/system/tasks/{task_id}")
    def fetch_system_task(self, response):
        """Get the background task by the given task Id.

        :param task_id:
        """
        return response

    # Get Version
    @get("/system/version")
    def fetch_version(self, response):
        """Get the current Learn server version."""
        return response

    # terms #

    # Get Term(s)
    @get("/terms/{term_id}")
    def fetch_terms(self, response):
        """Return a list of terms. / Load a term.

        :param term_id: The term ID.
        """
        return response

    # uploads #

    # users #

    # Get Users(s)
    @get("/users/{user_id}")
    def fetch_users(self, response):
        """Return a list of users. / Load a user.

        Properties returned will depend on the caller's entitlements.

        :param user_id: The user ID.
        """
        return response

    # Get Users Avatar
    # TODO: Handle this API call (HTTP header)
    @get("/users/{user_id}/avatar")
    def fetch_avatar(self, response):
        """Get a user avatar image.

        The response is an HTTP redirect rather than image raw data.
        It is up to the caller of the api to follow the redirect and download the image.

        Not yet implemented.
        :param user_id: The user ID.
        """
        raise NotImplementedError("API has not implemented this method yet.")
        return response

    # Get Observees
    @get("/users/{user_id}/observees")
    def fetch_observees(self, response):
        """Return a list of users being observed by a given user.

        :param user_id: The user ID.
        """
        return response

    # Get Observers
    @get("/users/{user_id}/observers")
    def fetch_observers(self, response):
        """Return a list of users observing a given user.

        :param user_id: The user ID.
        """
        return response

    # Get Current Active User By Id
    @get("/users/{user_id}/sessions")
    def fetch_current_active_user(self, response):
        """Display active session information for a specific user.

        :param user_id: The user ID.
        """
        return response

    @property
    def base_url(self) -> str:
        """API base URL."""
        return self._base_url

    @property
    def logger(self):
        """Logger for Blackboard API, set at level DEBUG."""
        return self._logger



"""BlackboardSync Tests"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pytest
from pydantic import ValidationError
from hypothesis import given, infer
from hypothesis import strategies as st
from pathvalidate import sanitize_filename

from blackboard_sync.blackboard import (BBFile, BBLink, BBCourse, BBLocale,
                                        BBDuration, BBAttachment, BBEnrollment,
                                        BBMembership, BBProctoring, BBResourceType,
                                        BBContentChild, BBCourseContent, BBContentHandler,
                                        BBAvailable, BBAvailability)

from .strategies import bb_unhandled_resource_type, bb_handled_resource_type, bb_resource_type


class TestBBCourseContent:
    def test_content_safe_title(self):
        og_title = "puppy.docx"
        sanitised = og_title

        cc = BBCourseContent(title=og_title)
        assert sanitised == cc.title_path_safe

    @pytest.mark.parametrize(
        ('title', 'sanitised'),
        (
            pytest.param("a/a\\a<a>a:a\"a|a?a*a.docx", f"{'a_' * 9}a.docx", id="unsafe chars"),
            pytest.param("../../../../important_file",
                         ".._.._.._.._important_file", id="path traversal unix"),
            pytest.param("..\\..\\..\\..\\important_file.exe",
                         ".._.._.._.._important_file.exe", id="path traversal windows")
        )
    )
    def test_content_sanitise_title(self, title, sanitised):
        obj = BBCourseContent(title=title)
        assert obj.title_path_safe == sanitised

    @given(st.text())
    def test_content_sanitise_title_hypothesis(self, filename):
        safe_path = sanitize_filename(filename or 'Title missing', replacement_text='_')
        obj = BBCourseContent(title=filename)
        assert safe_path or '' == obj.title_path_safe

    @given(bb_resource_type())
    def test_content_handler(self, res_type: str):
        assert BBContentHandler(id=res_type).id in (res_type.split('/')[-1], BBResourceType.other)

    @given(bb_handled_resource_type())
    def test_content_handler_handled(self, res_type: str):
        assert not (BBContentHandler(id=res_type).is_not_handled)

    @given(bb_unhandled_resource_type())
    def test_content_handler_unhandled(self, res_type: str):
        assert BBContentHandler(id=res_type).is_not_handled

    @pytest.mark.parametrize('available', ('Yes', 'No', 'Disabled'))
    def test_bb_available_not_other(self, available: str):
        assert BBAvailable(available) != BBAvailable.Other

    @pytest.mark.parametrize('available', ('UnexpectedValue', 'NotReal'))
    def test_bb_available_other(self, available: str):
        assert BBAvailable(available) == BBAvailable.Other

    @pytest.mark.parametrize('available', ('Yes', 'Term', 'PartiallyVisible' 'UnexpectedValue'))
    def test_bb_available_yes(self, available: str):
        assert BBAvailable(available)

    @pytest.mark.parametrize('available', ('No', 'Disabled'))
    def test_bb_available_no(self, available: str):
        assert not BBAvailable(available)

    @pytest.mark.parametrize('available', (True, False))
    def test_bb_available_bool(self, available: bool):
        with pytest.raises(ValidationError):
            BBAvailability(available=available)



"""
Checks for updates in GitHub.
"""

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import requests
from pathlib import Path
from typing import Optional
from packaging import version
from .__about__ import __version__


def is_inside_container() -> bool:
    return Path('/.flatpak-info').exists()

def check_for_updates() -> Optional[str]:
    """Checks if there is a newer release than the current on Github."""
    url = 'https://api.github.com/repos/jacobszpz/BlackboardSync/releases/latest'
    response = requests.get(url)
    if response.status_code == 200:
        json_response = response.json()
        tag = json_response['tag_name']
        tag = tag[1:] if tag[0] == 'v' else tag
        if version.parse(tag) > version.parse(__version__):
            return 'container' if is_inside_container() else json_response['html_url']

    return None





"""
BlackboardSync configuration manager
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import configparser
from typing import Any, Optional
from pathlib import Path
from functools import wraps
from datetime import datetime
from collections.abc import Callable

from appdirs import user_config_dir

from .__about__ import __author__


class Config(configparser.ConfigParser):
    """Base configuration manager class, which wraps a ConfigParser."""

    _logger = logging.getLogger(__name__)

    def __init__(self, config_file: Path, *args, **kwargs):
        super().__init__(converters={'path': Path, 'date': datetime.fromisoformat},
                         interpolation=None, *args, **kwargs)
        self._config_file = config_file
        self.read(self._config_file)

        # Set up logging
        self._logger.setLevel(logging.WARN)
        self._logger.addHandler(logging.StreamHandler())

    def save(self) -> None:
        """Save the current configuration to disk."""
        with self._config_file.open('w') as config_file:
            self.write(config_file)

    @staticmethod
    def persist(func) -> Callable[[Any, Any], None]:
        """A decorator to save any changes to the field to disk."""
        @wraps(func)
        def save_wrapper(self, *args: Any, **kwargs: Any) -> None:
            func(self, *args, **kwargs)
            self.save()
            self._logger.info("Updated configuration file")
        return save_wrapper

    @property
    def logger(self) -> logging.Logger:
        """Logger for BlackboardSync Configuration, set at level WARN."""
        return self._logger


class SyncConfig(Config):
    """Configuration manager for BlackboardSync."""
    _config_filename = "blackboard_sync"

    def __init__(self, custom_dir=None):
        config_dir = custom_dir or Path(user_config_dir(appauthor=__author__, roaming=True))
        super().__init__(config_dir / self._config_filename, empty_lines_in_values=False)

        if 'Sync' not in self:
            self['Sync'] = {}

        self._sync = self['Sync']

    @property
    def last_sync_time(self) -> Optional[datetime]:
        return self._sync.getdate('last_sync_time')

    @last_sync_time.setter
    @Config.persist
    def last_sync_time(self, last: Optional[datetime]) -> None:
        if last is None:
            self.remove_option('Sync', 'last_sync_time')
        else:
            self._sync['last_sync_time'] = last.isoformat()

    @property
    def download_location(self) -> Optional[Path]:
        # Default download location
        default = Path(Path.home(), 'Downloads', 'BlackboardSync')
        return self._sync.getpath('download_location') or default

    @download_location.setter
    @Config.persist
    def download_location(self, sync_dir: Path) -> None:
        self._sync['download_location'] = str(sync_dir)

    @property
    def university_index(self) -> Optional[int]:
        return self._sync.getint('university')

    @university_index.setter
    @Config.persist
    def university_index(self, university: int) -> None:
        self._sync['university'] = str(university)

    @property
    def min_year(self) -> Optional[int]:
        return self._sync.getint('min_year')

    @min_year.setter
    @Config.persist
    def min_year(self, year: Optional[int])-> None:
        self._sync['min_year'] = str(year or 0)



"""
Ping all the Blackboard API servers listed in universities.json

This script is not meant to run on CI or automatically since some Blackboard
Learn servers may be temporarily down at any given time.
Instead this script may be run manually to test the `universities.json` file
and that all API URLs are correct.

You may invoke this script with the following command from the project root
`pytest scripts/fetch_api_versions.py -s`
"""

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import requests
from enum import Enum, auto
import concurrent.futures
from json import JSONDecodeError

from blackboard_sync.institutions import _institutions

class APIResult(Enum):
    """Represents a specific outcome of an API request"""

    OK = auto()
    FAIL = auto()
    JSONERR = auto()
    CONNERR = auto()

    def __str__(self):
        color = lambda s: f"\033{s}\033[0m" 
        check, cross = color("[1;32m✓"), color("[91m✗")
        msg = {
            APIResult.OK: check,
            APIResult.FAIL: cross,
            APIResult.JSONERR: "[JSONDecodeError]",
            APIResult.CONNERR: "[ConnectionError]"
        }
        return msg[self]


def fetch_url(url, timeout):
    """Task that runs for every API url and attempts to fetch server version"""
    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()  # Raise an exception for non-2xx HTTP status codes
        data = response.json()
        if "learn" in data and "major" in data["learn"]:
            return APIResult.OK
        else:
            return APIResult.FAIL
    except requests.exceptions.ConnectionError:
        return APIResult.CONNERR
    except JSONDecodeError:
        return APIResult.JSONERR

def test_fetch_api_versions():
    """Ping many Blackboard API servers concurrently"""
    print("\n\n")
    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        v = lambda url: f"{url}learn/api/public/v1/system/version"
        # Start the load operations and mark each future with its name
        future_to_uni = {executor.submit(fetch_url, v(uni.api_url), 20): uni.name for uni in _institutions}

        for future in concurrent.futures.as_completed(future_to_uni):
            uni_name = future_to_uni[future]
            try:
                result = future.result()
            except Exception as exc:
                print('%r generated an exception: %s' % (uni_name, exc))
            else:
                print(f"{uni_name} {result}")


if __name__ == "__main__":
    raise SystemExit(test_fetch_api_versions())



#!/usr/bin/env python3

"""
BlackboardDownload,
mass download all user content from Blackboard
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import platform
from requests.exceptions import RequestException
from pathlib import Path
from getpass import getpass
from typing import Optional
from dateutil.parser import parse
from datetime import datetime, timezone
from pathvalidate import sanitize_filename
from concurrent.futures import ThreadPoolExecutor

from .blackboard import BlackboardSession, BBCourseContent, BBResourceType
from .webdav import ContentParser, Link, validate_webdav_response


class BlackboardDownload:
    """Blackboard download job."""

    _last_downloaded = datetime.fromtimestamp(0, tz=timezone.utc)

    _logger = logging.getLogger(__name__)
    _logger.setLevel(logging.DEBUG)
    _logger.addHandler(logging.StreamHandler())

    def __init__(self, sess: BlackboardSession,
                 download_location: Path,
                 last_downloaded: Optional[datetime] = None,
                 data_sources: list[str] = [],
                 min_year: Optional[int] = None):
        """BlackboardDownload constructor

        Download all files in blackboard recursively to download_location,
        only if they have been altered since specified datetime

        Keyword arguments:

        :param BlackboardSession sess: UCLan BB user session
        :param (str / Path) download_location: Where files will be stored
        :param str last_downloaded: Files modified before this will not be downloaded
        :param data_sources: List of valid data sources
        :param min_year: Only courses created on or after this year will be downloaded
        """

        self._sess = sess
        self._user_id = sess.username
        self._download_location = download_location
        self._data_sources = data_sources
        self._min_year = min_year
        self._files_processed = 0
        self.executor = ThreadPoolExecutor(max_workers=8)
        self.cancelled = False

        if last_downloaded is not None:
            self._last_downloaded = last_downloaded

        if not self.download_location.exists():
            self.download_location.mkdir(parents=True)
            self.logger.info("Created download folder")

    def _create_desktop_link(self, path: Path, url: str, comment: str = "") -> None:
        """Creates a platform-aware internet shortcut"""
        if platform.system() in ["Windows", "Darwin"]:
            contents = f"[InternetShortcut]\nURL={url}"
            path = path.with_suffix(".url")
        else:
            contents = f"[Desktop Entry]\nIcon=text-html\nType=Link\nURL[$e]={url}"

        with path.open("w") as f:
            f.write(contents)
            self.logger.info(f"Created internet link file at {path}")

    def _download_file(self, course_id: str, content_id: str, attachment_id: str, file_path: Path) -> None:
        """Get stream for blackboard file and download."""
        try:
            d_stream = self._sess.download(course_id=course_id,
                                       content_id=content_id,
                                       attachment_id=attachment_id)
            self._download_stream(d_stream, file_path)
        except RequestException as e:
            self.logger.warn(f"Error while downloading file {content_id}")

    def _download_webdav_file(self, link: str, file_path: Path) -> None:
        try:
            response = self._sess.download_webdav(webdav_url=link)
            if validate_webdav_response(response, link, self._sess.base_url):
                self._download_stream(response, file_path)
            else:
                self.logger.info(f"Not downloading webdav/ext file {link}")
        except RequestException as e:
            self.logger.warn(f"Error while downloading webdav/ext file {link}")


    def _download_stream(self, stream, file_path):
        """Generic stream download function."""
        with file_path.open("wb") as f:
            self.logger.info(f"Writing to {file_path}")
            for chunk in stream.iter_content(chunk_size=1024):
                f.write(chunk)

    def _handle_file(self, content: BBCourseContent, parent_path: Path,
                     course_id: str, depth: int = 0) -> None:
        """Download BBContent recursively, depending on filetype"""
        if self.cancelled:
            return

        res = content.contentHandler

        handler_id = res.id if res is not None else '?'
        self.logger.info(f"{'    ' * depth}{content.title}[{handler_id}]")

        body_path = parent_path
        file_path = Path(parent_path, content.title_path_safe)
        has_changed = True

        if content.modified is not None:
            has_changed = (content.modified >= self._last_downloaded)

        if res is None:
            pass

        elif res == BBResourceType.folder:
            try:
                body_path = file_path
                children = self._sess.fetch_content_children(course_id=course_id,
                                                             content_id=content.id)

                if children or content.body:
                    file_path.mkdir(exist_ok=True, parents=True)

                for child in children:
                    if child.contentHandler is not None:
                        self._handle_file(child, file_path, course_id, depth + 1)
            except ValueError:
                self.logger.warn(f"Error while getting children for {course_id}")

        # Omit file if it hasn't been modified since last sync
        elif res in (BBResourceType.file, BBResourceType.document) and has_changed:
            attachments = []

            try:
                attachments = self._sess.fetch_file_attachments(course_id=course_id,
                                                                content_id=content.id)

            except RequestException:
                self.logger.warn(f"Net error while getting attachments for {course_id}")
            except ValueError:
                self.logger.warn(f"Error while getting attachments for {course_id}")

            if len(attachments) > 1:
                file_path.mkdir(exist_ok=True, parents=True)
                body_path = file_path

            for attachment in attachments:
                download_path = Path(body_path / attachment.fileName)
                if attachment.mimeType.startswith('video/'):
                    self.logger.info(f'Not downloading {attachment.fileName}')
                else:
                    if not self.cancelled and content.id is not None:
                        self.executor.submit(self._download_file,
                                             course_id, content.id,
                                             attachment.id, download_path)

        elif res == BBResourceType.externallink and has_changed:
            # Place link under folder of its own, in case it has a body
            file_path.mkdir(exist_ok=True, parents=True)
            link_path = file_path / content.title_path_safe
            if res.url is not None:
                self._create_desktop_link(link_path, res.url)

        elif not res.is_not_handled and has_changed:
            self.logger.warning(f"Not handled, {content.title}")

        # If item has body, write in markdown file
        if content.body and has_changed:
            file_path.mkdir(exist_ok=True, parents=True)

            # Parse content.body for more attachments
            parser = ContentParser(content.body, self._sess.base_url)

            for body_link in parser.links:
                safe_title = sanitize_filename(body_link.text, replacement_text='_')
                download_path = Path(file_path / safe_title)
                self.executor.submit(self._download_webdav_file, body_link.href, download_path)

            with Path(file_path, f"{content.title_path_safe}.html").open('w', encoding='utf-8') as html_content:
                html_content.write(parser.body)

    def download(self) -> Optional[datetime]:
        """Retrieve the user's courses, and start download of all contents

        :return: Datetime when method was called.
        """
        if self.cancelled:
            return None

        start_time = datetime.now(timezone.utc)

        self.logger.info("Fetching user memberships")

        memberships = self._sess.fetch_user_memberships(user_id=self.user_id)

        # Filter courses by data source
        if self._data_sources:
            memberships = [m for m in memberships if m.dataSourceId in self._data_sources]

        # Filter courses by creation year
        if self._min_year is not None:
            memberships = [m for m in memberships if m.created.year >= self._min_year]

        for ms in memberships:
            if self.cancelled:
                break

            private = False
            self.logger.debug("Fetching course")
            try:
                course = self._sess.fetch_courses(course_id=ms.courseId)
            except ValueError as e:
                if not (private := (str(e) == 'Private course')):
                    raise e

            if not private:
                self.logger.info(f"<{course.code}> - <{course.title}>")

                course_contents = self._sess.fetch_contents(course_id=course.id)

                if course_contents:
                    course_path = Path(self.download_location / str(ms.created.year) / course.title)

                for content in course_contents:

                    if self.cancelled:
                        break

                    self._handle_file(content, course_path, course.id, 1)
        self.executor.shutdown(wait=True, cancel_futures=self.cancelled)

        if self.cancelled:
            return None
        return start_time

    def cancel(self) -> None:
        """Cancel the download job."""
        self.cancelled = True

    @property
    def download_location(self) -> Path:
        """The location where files will be downloaded to."""
        return self._download_location

    @property
    def data_sources(self) -> list[str]:
        """Filter for courses."""
        return self._data_sources

    @data_sources.setter
    def data_sources(self, sources: list[str]) -> None:
        self._data_sources = sources

    @property
    def user_id(self) -> str:
        """User ID used for API calls."""
        return self._user_id

    @property
    def files_processed(self) -> int:
        """Number of files that have been downloaded."""
        return self._files_processed

    @property
    def logger(self) -> logging.Logger:
        """Logger for BlackboardDownload, set at level DEBUG."""
        return self._logger




"""BlackboardSync Qt GUI."""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import json
import platform
import subprocess
import webbrowser
from enum import IntEnum
from typing import Optional
from pathlib import Path
from datetime import datetime

from PyQt5 import uic
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtCore import Qt, QSettings, QUrl, pyqtSlot, pyqtSignal
from PyQt5.QtNetwork import QNetworkCookie
from PyQt5.QtWidgets import (QMenu, QStyle, QAction, QDialog, QWidget, QWizard,
                             QCompleter, QFileDialog, QMessageBox, QApplication,
                             QSystemTrayIcon, QComboBox, QLabel, QCheckBox, QSpinBox)
from requests.cookies import RequestsCookieJar
from PyQt5.QtWebEngineCore import QWebEngineCookieStore
from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEnginePage, QWebEngineProfile


class SyncPeriod(IntEnum):
    """Enum containing all valid Sync intervals for this UI."""

    HALF_HOUR = 60 * 30
    ONE_HOUR = 60 * 60
    SIX_HOURS = 60 * 60 * 6


class Assets:
    """Helper class to get the path of app assets."""

    _icon_filename = 'logo.png'
    _watermark_filename = 'watermark.png'

    @staticmethod
    def _get_qt_asset_path(asset_file) -> Path:
        """Get the `Path` corresponding to a Qt UI file."""
        return (Path(__file__).parent / f"{asset_file}.ui").resolve()

    @classmethod
    def load_ui(cls, qt_obj):
        """Load a UI file for a `QObject`."""
        uic.loadUi(cls._get_qt_asset_path(qt_obj.__class__.__name__), qt_obj)

    @staticmethod
    def _get_asset_path(icon) -> Path:
        """Get the `Path` of a media asset."""
        return (Path(__file__).parent.parent / 'assets' / icon).resolve()

    @classmethod
    def icon(cls) -> QIcon:
        """`QIcon` of application logo."""
        return QIcon(str(cls._get_asset_path(cls._icon_filename)))

    @classmethod
    def watermark(cls) -> QPixmap:
        """`QPixmap` of application watermark."""
        wm = QPixmap(str(cls._get_asset_path(cls._watermark_filename)))
        wm = wm.scaledToWidth(100)
        return wm


class OSUtils:
    @staticmethod
    def open_dir_in_file_browser(dir_to_open: Path) -> None:
        """Start a subprocess to open the default file explorer at the given location."""
        if sys.platform == "win32":
            os.startfile(dir_to_open)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", dir_to_open])
        else:
            subprocess.Popen(["xdg-open", dir_to_open])

    @staticmethod
    def add_to_startup() -> None:
        """Add the app to start up on macOS."""
        if platform.system() != "Darwin":
            return

        # Set the paths and filenames
        app_path = '/Applications/BBSync.app'
        launch_dir = Path("~/Library/LaunchAgents").expanduser()

        if not launch_dir.exists():
            launch_dir.mkdir()

        plist_path = launch_dir / "app.bbsync.plist"
        plist_path.touch()

        # Create the QSettings object
        settings = QSettings(str(plist_path), QSettings.NativeFormat)

        # Set the launch agent properties
        settings.setValue('Label', 'app.bbsync.BBSync')
        settings.setValue('ProgramArguments', app_path)
        settings.setValue('RunAtLoad', True)
        settings.setValue('KeepAlive', False)

        # Save the settings to create the plist file
        settings.sync()


class SyncTrayMenu(QMenu):
    """`QMenu` associated with app system tray icon."""

    _unauthenticated_status = "You haven't logged in"

    def __init__(self, logged_in: bool = False, last_synced: str = ""):
        """Create the menu for a `SyncTrayIcon`.

        :param bool logged_in: Whether user is currently logged in.
        :param str last_synced: Last sync time shown.
        """
        super().__init__()
        self._last_synced = ""

        self._init_ui()
        self.update_last_synced(last_synced)
        self.set_logged_in(logged_in)

    def _init_ui(self) -> None:
        sync_icon = QApplication.style().standardIcon(QStyle.SP_BrowserReload)
        close_icon = QApplication.style().standardIcon(QStyle.SP_DialogCloseButton)
        open_dir_icon = QApplication.style().standardIcon(QStyle.SP_DirOpenIcon)

        self.refresh = QAction("Sync now")
        self.refresh.setIcon(sync_icon)
        self.addAction(self.refresh)

        self.open_dir = QAction("Open downloads")
        self.open_dir.setIcon(open_dir_icon)
        self.addAction(self.open_dir)

        self.preferences = QAction("Preferences")
        self.addAction(self.preferences)

        self.addSeparator()

        self._status = QAction(self._unauthenticated_status)
        self._status.setEnabled(False)
        self.addAction(self._status)

        self.log_in = QAction("Log In")
        self.addAction(self.log_in)

        self.reset_setup = QAction("Redo Setup")
        self.addAction(self.reset_setup)

        self.quit = QAction("Quit")
        self.quit.setIcon(close_icon)
        self.addAction(self.quit)

    def set_logged_in(self, logged: bool) -> None:
        """Set the UI to reflect logged-in status."""
        self.refresh.setVisible(logged)
        self.preferences.setVisible(logged)
        self.reset_setup.setVisible(not logged)
        self.open_dir.setVisible(logged)
        self.log_in.setVisible(not logged)

        if logged:
            self._status.setText(f"Last Synced: {self._last_synced}")
        else:
            self._status.setText("Not Logged In")

    def update_last_synced(self, last: str) -> None:
        """Update the time of last download shown to user."""
        self._last_synced = last
        self._status.setText(f"Last Synced: {self._last_synced}")

    def toggle_currently_syncing(self, syncing: bool) -> None:
        """Toggle the currently syncing indicator."""
        self.refresh.setEnabled(not syncing)

        if syncing:
            self._status.setText("Downloading now...")


class SyncTrayIcon(QSystemTrayIcon):
    """BlackboardSync system tray icon."""

    _tooltip = "Blackboard Sync"
    _sync_signal = pyqtSignal()
    _login_signal = pyqtSignal()
    _settings_signal = pyqtSignal()
    _reset_setup_signal = pyqtSignal()
    _quit_signal = pyqtSignal()
    _open_dir_signal = pyqtSignal()
    _show_menu_signal = pyqtSignal()

    def __init__(self):
        """Create a `QSystemTrayIcon`."""
        super().__init__()
        self._init_ui()

    def _init_ui(self) -> None:
        # Create the icon
        icon = Assets.icon()

        # Create the tray
        self.setIcon(icon)
        self.setVisible(True)

        # Create the menu
        self._menu = SyncTrayMenu()

        self._sync_signal = self._menu.refresh.triggered
        self._login_signal = self._menu.log_in.triggered
        self._settings_signal = self._menu.preferences.triggered
        self._reset_setup_signal = self._menu.reset_setup.triggered
        self._quit_signal = self._menu.quit.triggered
        self._open_dir_signal = self._menu.open_dir.triggered
        self._show_menu_signal = self._menu.aboutToShow

        # Add the menu to the tray
        self.setContextMenu(self._menu)
        self.setToolTip(self._tooltip)

    def set_logged_in(self, logged: bool) -> None:
        """Set logged-in status in menu."""
        self._menu.set_logged_in(logged)

    def update_last_synced(self, last: str) -> None:
        """Update last sync time in menu."""
        self._menu.update_last_synced(last)

    def toggle_currently_syncing(self, syncing: bool) -> None:
        """Toggle currently syncing indicator in menu."""
        self._menu.toggle_currently_syncing(syncing)

    def show_msg(self, title: str, msg: str, severity: int = 1, duration: int = 10) -> None:
        """Show the user a message through the tray icon."""
        icons = { 0: QSystemTrayIcon.NoIcon,
                  1: QSystemTrayIcon.Information,
                  2: QSystemTrayIcon.Warning,
                  3: QSystemTrayIcon.Critical }
        duration = duration * 1000
        severity = 0 if severity < 0 or severity > 3 else severity
        self.showMessage(title, msg, icons[severity], duration)

    @property
    def sync_signal(self):
        """Fire if user forces sync."""
        return self._sync_signal

    @property
    def login_signal(self):
        """Fire once user is authenticated."""
        return self._login_signal

    @property
    def settings_signal(self):
        """Fire when the settings menu is opened."""
        return self._settings_signal

    @property
    def reset_setup_signal(self):
        """Fire when the user wants to reset the initial setup."""
        return self._reset_setup_signal

    @property
    def quit_signal(self):
        """Fire once user decides to quit app."""
        return self._quit_signal

    @property
    def open_dir_signal(self):
        """Fire once user wants to open download directory."""
        return self._open_dir_signal

    @property
    def show_menu_signal(self):
        """Fire when menu is about to be shown."""
        return self._show_menu_signal


class RedownloadDialog(QMessageBox):
    """`QMessageBox` shown after a change in download location.

    It consults the user about whether files should be redownloaded to
    the new location or not.
    """

    _window_title = "Redownload all files?"
    _dialog_text = "Should BlackboardSync redownload all files to the new location?"
    _info_text = "Answer no if you intend to move all past downloads manually (Recommended)"

    def __init__(self):
        """Create a `RedownloadDialog`."""
        super().__init__()
        self._init_ui()

    def _init_ui(self) -> None:
        self.setText(self._dialog_text)
        self.setInformativeText(self._info_text)
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.setDefaultButton(QMessageBox.No)
        self.setWindowTitle(self._window_title)
        self.setIcon(QMessageBox.Question)
        self.setWindowIcon(Assets.icon())

    @property
    def redownload(self) -> bool:
        """Indicate if files have to be redownloaded."""
        return self.exec() == QMessageBox.Yes


class UpdateFoundDialog(QMessageBox):
    """`QMessageBox` shown after a more recent version was found."""

    _window_title = "New BlackboardSync release available"
    _dialog_text = "A new version of BlackboardSync is now available!"
    _info_text = "Please download the latest version from the official GitHub repository"

    def __init__(self):
        """Create a `UpdateFoundDialog`."""
        super().__init__()
        self._init_ui()

    def _init_ui(self) -> None:
        self.setText(self._dialog_text)
        self.setInformativeText(self._info_text)
        self.setStandardButtons(QMessageBox.Open | QMessageBox.Cancel)
        self.setDefaultButton(QMessageBox.Open)
        self.setWindowTitle(self._window_title)
        self.setIcon(QMessageBox.Information)
        self.setWindowIcon(Assets.icon())

    @property
    def should_update(self) -> bool:
        """Indicate if BBSync should be updated."""
        return self.exec() == QMessageBox.Open


class SettingsWindow(QWidget):
    """Settings windown UI element."""

    _window_title = "Settings"
    _initial_position = (300, 300)
    _log_out_signal = pyqtSignal()
    _setup_wiz_signal = pyqtSignal()
    _save_signal = pyqtSignal()

    def __init__(self):
        """Create instance of SettingsWindow."""
        super().__init__()
        self.frequency_combo: QComboBox
        self.current_session_label: QLabel
        self.download_location_hint: QLabel
        self._init_ui()

    def _init_ui(self):
        Assets.load_ui(self)

        self.move(*self._initial_position)
        self.setWindowTitle(self._window_title)

        self.select_download_location.clicked.connect(self._choose_location)
        self._log_out_signal = self.log_out_button.clicked
        self._setup_wiz_signal = self.setup_button.clicked
        self._save_signal = self.button_box.accepted

    def _choose_location(self) -> None:
        if (location := self._file_chooser_dialog()):
            self.download_location = location

    def _file_chooser_dialog(self) -> Optional[Path]:
        self.file_chooser = QFileDialog()
        self.file_chooser.setFileMode(QFileDialog.Directory)

        if self.file_chooser.exec():
            new_location = self.file_chooser.directory()
            return Path(new_location.path())

        return None

    @property
    def download_location(self) -> Path:
        """`Path` of download location."""
        return self._download_location

    @download_location.setter
    def download_location(self, location: Path) -> None:
        self._download_location = location.resolve()
        self.download_location_hint.setText(str(self._download_location))

    @property
    def sync_frequency(self) -> int:
        """Seconds to wait between each sync job."""
        return int([*SyncPeriod][self.frequency_combo.currentIndex()])

    @sync_frequency.setter
    def sync_frequency(self, f: int) -> None:
        self.frequency_combo.setCurrentIndex([*SyncPeriod].index(SyncPeriod(f)))

    @property
    def username(self) -> str:
        """Username of current session."""
        return self.current_session_label.text()

    @username.setter
    def username(self, username: str) -> None:
        if username:
            self.current_session_label.setText(f"Logged in as {username}")
        else:
            self.current_session_label.setText("Not currently logged in")

    @property
    def log_out_signal(self):
        """Fire when user chooses to log out."""
        return self._log_out_signal

    @property
    def setup_wiz_signal(self):
        """Fire when user wants to redo initial setup."""
        return self._setup_wiz_signal

    @property
    def save_signal(self):
        """Fire when settings are saved."""
        return self._save_signal


class LoginWebView(QWidget):
    """Blackboard login widget."""

    _login_complete_signal = pyqtSignal()

    def __init__(self, start_url: str, target_url: str):
        """Create instance of LoginWebView."""
        super().__init__()
        self.start_url = start_url
        self.target_url = target_url

        self.web_view : QWebEngineView
        self._init_ui()
        self._cookie_jar = RequestsCookieJar()

    def _init_ui(self) -> None:
        Assets.load_ui(self)
        self.web_view.load(QUrl.fromUserInput(self.start_url))
        self.web_view.loadFinished.connect(self._page_load_handler)
        self.done_button.clicked.connect(self._login_complete_signal)
        self._cookie_store.cookieAdded.connect(self._cookie_added_handler)

    def _page_load_handler(self) -> None:
        if self.url.startswith(self.target_url):
            self._login_complete_signal.emit()

    def _cookie_added_handler(self, cookie: QNetworkCookie) -> None:
        self._cookie_jar.set(
            cookie.name().data().decode(),
            cookie.value().data().decode(),
            domain=cookie.domain(),
            path=cookie.path(),
            secure=cookie.isSecure()
        )

    def restore(self) -> None:
        self.web_view.setPage(None)
        self.clear_cookie_store()
        self.web_view.load(QUrl.fromUserInput(self.start_url))

    def clear_cookie_store(self) -> None:
        """Clear the HTTP cache and cookies."""
        self._cookie_store.deleteAllCookies()
        self._engine_profile.clearHttpCache()

    @property
    def url(self) -> str:
        """URL of current website."""
        return self.web_view.url().toString()

    @property
    def cookie_jar(self) -> RequestsCookieJar:
        """Contains session cookies of the current session."""
        return self._cookie_jar

    @property
    def _engine_page(self) -> QWebEnginePage:
        return self.web_view.page()

    @property
    def _engine_profile(self) -> QWebEngineProfile:
        return self._engine_page.profile()

    @property
    def _cookie_store(self) -> QWebEngineCookieStore:
        return self._engine_profile.cookieStore()

    @property
    def login_complete_signal(self):
        """Fire when the login flow has completed."""
        return self._login_complete_signal


class SetupWizard(QWizard):
    """Initial setup wizard."""

    class Pages(IntEnum):
        """Pages contained in the wizard."""

        INTRO = 0
        INSTITUTION = 1
        DOWNLOAD_LOCATION = 2
        DOWNLOAD_SINCE = 3
        LAST = 3

    _help_website = 'https://github.com/sanjacob/BlackboardSync'

    def __init__(self, institutions: list[str]):
        """Create a `SetupWizard`.

        :param list[str] institutions: List of institution names
        """
        super().__init__()

        self.uni_selection_box: QComboBox
        self.since_all_checkbox: QCheckBox
        self.date_spinbox: QSpinBox
        self.institutions = institutions
        self._init_ui()
        self._has_chosen_location = False

    def _init_ui(self):
        Assets.load_ui(self)
        self.uni_selection_box.addItems(self.institutions)
        self.uni_selection_box.clearEditText()

        self.completer = QCompleter(self.institutions, self)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setFilterMode(Qt.MatchContains)
        self.uni_selection_box.setCompleter(self.completer)

        self.file_chooser = QFileDialog()
        self.file_chooser.setFileMode(QFileDialog.Directory)
        self.sync_location_button.clicked.connect(self._choose_location)

        year = datetime.today().year
        self.date_spinbox.setRange(2000, year)
        self.date_spinbox.setValue(year)
        self.date_spinbox.setEnabled(False)

        self.since_all_checkbox.stateChanged.connect(
            lambda state: self.date_spinbox.setEnabled(state != Qt.Checked)
        )

        self.uni_selection_page.registerField(
            "userInstitution*",
            self.uni_selection_box.lineEdit()
        )

        self.sync_location_page.registerField(
            "syncLocation*",
            self.sync_location_button,
            property="text",
            changedSignal=self.sync_location_button.clicked
        )

        self.intro_page.setPixmap(QWizard.WatermarkPixmap,
                                  Assets.watermark())

    def initializePage(self, id) -> None:
        if id == self.Pages.DOWNLOAD_LOCATION:
            if self._has_chosen_location:
                self._set_location()

    def validateCurrentPage(self) -> bool:
        """Override QWizard method to validate pages."""
        id = self.currentId()
        valid = True

        if id == self.Pages.INSTITUTION:
            # Do not move forward if institution is not recognised
            if not self._institution_is_valid():
                self._show_not_supported_dialog()
                valid = False

        return valid

    def _set_location(self):
        dir = self.download_location.name or str(self.download_location)
        self.sync_location_button.setText(dir)

    def _choose_location(self):
        if self.file_chooser.exec():
            self._set_location()
            self._has_chosen_location = True

    def _show_not_supported_dialog(self):
        error_dialog = UniNotSupportedDialog(self._help_website)
        error_dialog.exec()

    def _institution_is_valid(self) -> bool:
        return self.field("userInstitution") == self.institution

    @property
    def institution(self) -> str:
        """Text of item selected in institution combo box."""
        return self.uni_selection_box.itemText(self.institution_index)

    @property
    def institution_index(self) -> int:
        """Index of item selected in institution combo box."""
        return self.uni_selection_box.currentIndex()

    @property
    def download_location(self) -> Path:
        """Sync location path selected by user."""
        return Path(self.file_chooser.directory().path())

    @property
    def min_year(self) -> Optional[int]:
        """Courses from this year onward will be downloaded."""
        if not self.since_all_checkbox.isChecked():
            return self.date_spinbox.value()
        return None


class UniNotSupportedDialog(QDialog):
    """`QDialog` about unsupported Blackboard partners."""

    def __init__(self, help_url: str):
        """Create instance of dialog.

        :param str help_url: URL to help website
        """
        super().__init__()
        self._init_ui()
        self._help_url = help_url

    def _init_ui(self):
        Assets.load_ui(self)
        self.button_box.helpRequested.connect(self._open_help_website)

    @pyqtSlot()
    def _open_help_website(self):
        webbrowser.open(self._help_url)



#!/usr/bin/env python3

"""
BlackboardSync
~~~~~~~~~~

Keep your Blackboard files synced locally

:copyright: (c) 2023, Jacob Sánchez Pérez.
:license: GPL v2.0, see LICENSE for details.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# Also available at https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

from .__about__ import __title__, __summary__, __uri__, __version__
from .__about__ import __author__, __email__, __license__
from .__about__ import __copyright__



"""BlackboardSync Tests"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch, ANY

import pytest
from hypothesis import given, assume
from hypothesis import strategies as st
from hypothesis import provisional as pr

from blackboard_sync.blackboard.api import BlackboardSession
from blackboard_sync.download import BlackboardDownload
from blackboard_sync.blackboard.blackboard import BBMembership, BBCourse


@pytest.fixture
def mock_course():
    mock = Mock(spec=BBCourse)
    mock.code = 'TEST_BBC_CODE'
    mock.title = 'TEST_BBC_TITLE'
    mock.id = 'TEST_BBC_ID'
    return mock

@pytest.fixture
def mock_membership(mock_course):
    mock = Mock(spec=BBMembership)
    mock.courseId = mock_course.id
    return mock

@pytest.fixture
def mock_session(mock_membership, mock_course):
    mock = Mock(spec=BlackboardSession)
    mock.username = "example"
    mock.fetch_user_memberships.return_value = [ mock_membership ]
    mock.fetch_courses.return_value = mock_course
    return mock


@patch('blackboard_sync.download.platform')
@given(url=pr.urls(), current_platform=st.sampled_from(['Windows', 'Darwin']))
def test_create_link_windows_darwin(url, current_platform, mock_platform):
    mock_session = Mock(spec=BlackboardSession)
    mock_session.username = 'example'

    # Patch platform value so it can run independently of current OS.
    mock_platform.system.return_value = current_platform

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        link_path = Path(tmp_path / "link")
        real_path = link_path.with_suffix('.url')
        assert not real_path.exists()

        download = BlackboardDownload(mock_session, tmp_path)
        download._create_desktop_link(link_path, url)
        assert real_path.exists()
        contents = f"[InternetShortcut]\nURL={url}"

        with real_path.open('r') as link_file:
            assert link_file.read() == contents

@patch('blackboard_sync.download.platform')
@given(url=pr.urls(), current_platform=st.sampled_from(['Linux', '']))
def test_create_link_default(url, current_platform, mock_platform):
    mock_session = Mock(spec=BlackboardSession)
    mock_session.username = 'example'

    # Patch platform value so it can run independently of current OS.
    mock_platform.system.return_value = current_platform

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        link_path = Path(tmp_path / "link")
        assert not link_path.exists()

        download = BlackboardDownload(mock_session, tmp_path)
        download._create_desktop_link(link_path, url)
        assert link_path.exists()

        contents = f"[Desktop Entry]\nIcon=text-html\nType=Link\nURL[$e]={url}"

        with link_path.open('r') as link_file:
            assert link_file.read() == contents

def test_download_location(mock_session, tmp_path):
    download = BlackboardDownload(mock_session, tmp_path)
    assert download.download_location == tmp_path
    assert tmp_path.exists()

def test_logger(mock_session, tmp_path):
    download = BlackboardDownload(mock_session, tmp_path)
    assert isinstance(download.logger, logging.Logger)

def test_download_method_call_fetch_user_memberships_with_username(mock_session, tmp_path):
    expected_user = "test_username"
    mock_session.fetch_user_memberships.return_value = []
    mock_session.username = expected_user
    download = BlackboardDownload(mock_session, tmp_path)
    download.download()
    mock_session.fetch_user_memberships.assert_called_once_with(user_id=expected_user)

def test_download_method_call_fetch_courses_skip_private(mock_session, tmp_path):
    expected_course_id = 'TEST_BBC_ID'
    mock_session.fetch_courses.side_effect = ValueError('Private course')
    download = BlackboardDownload(mock_session, tmp_path)
    download.download()
    mock_session.fetch_courses.assert_called_once_with(course_id=expected_course_id)

def test_download_method_call_fetch_courses_raise_error(mock_session, tmp_path):
    mock_session.fetch_courses.side_effect = ValueError('Other error')
    download = BlackboardDownload(mock_session, tmp_path)
    with pytest.raises(ValueError) as excinfo:
        download.download()
    assert str(excinfo.value) == 'Other error'

def test_download_method_call_fetch_contents_with_id(mock_session, tmp_path):
    mock_session.fetch_contents.return_value = []
    download = BlackboardDownload(mock_session, tmp_path)
    download.download()
    mock_session.fetch_contents.assert_called_once_with(course_id='TEST_BBC_ID')



#!/usr/bin/env python3

"""BlackboardSync API Tests"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pytest
import requests

from blackboard_sync.blackboard.api import SafeFormat, BlackboardSession

# SafeFormat

def test_missing():
    safe_dict = SafeFormat({'a': 1})
    assert safe_dict['b'] == ''

def test_present():
    safe_dict = SafeFormat({'a': 1})
    assert safe_dict['a'] == 1


# BlackboardSession

@pytest.fixture
def mock_bbsession(mocker):
    """Fixture for the MockBlackboardSession class."""
    class MockBlackboardSession:
        _base_url = 'https://api.example.com'
        _timeout = 1
        # Logger
        logger = mocker.Mock()
        # Requests mocks
        response_mock = mocker.Mock()
        response_mock.json.return_value = {}

        session_mock = mocker.Mock()
        session_mock.get.return_value = response_mock
        _bb_session = session_mock

        def __init__(self, return_value):
            self.response_mock.json.return_value = return_value

        @BlackboardSession.get('/test_endpoint/{optional_id}')
        def api_operation(self, response):
            return response

    return MockBlackboardSession


@pytest.mark.parametrize("expected_error, response_data", [
    ('Server response empty', {}),
    ('Not authorized', {'status': 401, 'message': 'Error message'}),
    ('Private course', {'status': 403, 'message': 'Error message', 'code': 'bb-rest-course-is-private'}),
    ('Server responded with an error code', {'status': 418, 'message': 'Error message'})
])
def test_get_decorator_errors(mock_bbsession, expected_error, response_data):
    with pytest.raises(ValueError) as excinfo:
        mock_bbsession(response_data).api_operation()
    assert str(excinfo.value) == expected_error


@pytest.mark.parametrize("results", [
    {'a': 1, 'b': 2, 'c': 'd'},
])
def test_get_decorator_response(mock_bbsession, results):
    s = mock_bbsession(results)
    api_response = s.api_operation()
    assert api_response == results


@pytest.mark.parametrize("results", [
    [],
    [1, 2, 3],
    ['a', 'b', 'c']
])
def test_get_decorator_results(mock_bbsession, results):
    s = mock_bbsession({'results': results})
    api_response = s.api_operation()
    assert api_response == results


def test_get_decorator_endpoint(mock_bbsession, mocker):
    s = mock_bbsession({'results': []})
    api_response = s.api_operation()
    s.session_mock.get.assert_called_once_with('https://api.example.com/learn/api/public/v1/test_endpoint', timeout=mocker.ANY, params=mocker.ANY)


def test_get_decorator_endpoint_with_id(mock_bbsession, mocker):
    s = mock_bbsession({'results': []})
    api_response = s.api_operation(optional_id='test_id')
    s.session_mock.get.assert_called_once_with('https://api.example.com/learn/api/public/v1/test_endpoint/test_id', timeout=mocker.ANY, params=mocker.ANY)


def test_get_decorator_params(mock_bbsession, mocker):
    s = mock_bbsession({'results': []})
    expected = {'paramA': 1, 'paramB': 2, 'paramC': 'hello world'}
    api_response = s.api_operation(**expected)
    s.session_mock.get.assert_called_once_with(mocker.ANY, timeout=mocker.ANY, params=expected)


def test_get_decorator_params_with_endpoint_id(mock_bbsession, mocker):
    s = mock_bbsession({'results': []})
    expected = {'paramA': 1, 'paramB': 2, 'paramC': 'hello world', 'optional_id': 'passed'}
    # Even API endpoint are passed as parameters, which should not be an issue
    # but should be noted regardless (the server will simply ignore them)
    api_response = s.api_operation(**expected)
    s.session_mock.get.assert_called_once_with('https://api.example.com/learn/api/public/v1/test_endpoint/passed', timeout=mocker.ANY, params=expected)

# monkeypatch.setattr(requests, "Session", lambda *args, **kwargs: MockSession())
# s = BlackboardSession('base', None)
# s.fetch_announcements()



"""
Blackboard API.

~~~~~~~~~~~~~~

A python interface for the Blackboard REST APIs.

Basic usage:
    >>> from blackboard import BlackboardSession
    >>> s = BlackboardSession(base_url='https://...', cookies=...)
    >>> s.fetch_version()


Full documentation available at <https://blackboardsync.readthedocs.io>.

:copyright: (c) 2021, Jacob Sánchez Pérez.
:license: GPL v2.0, see LICENSE for more details.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# Also available at https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

from .api import BlackboardSession
from .blackboard import (BBFile, BBLink, BBCourse, BBLocale, BBDuration,
                         BBAttachment, BBEnrollment, BBMembership,
                         BBProctoring, BBAvailability, BBContentChild,
                         BBCourseContent, BBContentHandler, BBResourceType,
                         BBAvailable)

__all__ = ['BBLocale', 'BBDuration', 'BBEnrollment', 'BBProctoring', 'BBFile',
           'BBAttachment', 'BBLink', 'BBContentHandler', 'BBAvailability',
           'BBCourseContent', 'BBContentChild', 'BBMembership', 'BBCourse',
           'BlackboardSession', 'BBResourceType', 'BBAvailable']



#!/usr/bin/env python3

"""BlackboardSync Controller."""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import webbrowser
from typing import Optional

from PyQt5.QtGui import QWindow
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QStyleFactory, QSystemTrayIcon, QWidget

from .sync import BlackboardSync
from .institutions import Institution, get_names, InstitutionLogin
from .__about__ import __title__, __version__
from .updates import check_for_updates
from .qt.qt_elements import (LoginWebView, SyncTrayIcon, SettingsWindow,
                             RedownloadDialog, OSUtils, SetupWizard, UpdateFoundDialog)


class BBSyncController:
    """Connects an instance of BlackboardSync with the UI module."""

    tray_msg = {
        "container_update": ('Updates available', 'You can update BlackboardSync from the Software Center', 1, 4),
        "download_started": ('The download has started', 'BlackboardSync is running in the background. Find it in the system tray.'),
        "download_error": ('The download cannot be completed', 'There was an error validating your course content. Please report this issue.', 3, 10)
    }

    def __init__(self):
        """Create an instance of the BlackboardSync Desktop App."""
        # Create model, which will try to retrieve existing configuration
        # If unsuccessful, we bring up login page
        self.model = BlackboardSync()

        self.app = QApplication(sys.argv)
        self.app.setApplicationName(__title__)
        self.app.setApplicationVersion(__version__)

        QApplication.setStyle(QStyleFactory.create("Fusion"))

        self._init_ui()
        self._has_notified_error = False

        if self.model.university is None:
            OSUtils.add_to_startup()
            self._show_setup_window()
        else:
            self._build_login_window(self.model.university.login)
            self._show_login_window()
        self.app.exec()

    def _init_ui(self) -> None:
        self.setup_window = SetupWizard(get_names())
        self.setup_window.accepted.connect(self._setup_complete)

        self.login_window : Optional[LoginWebView] = None

        self.config_window = SettingsWindow()

        self.config_window.log_out_signal.connect(self._log_out)
        self.config_window.setup_wiz_signal.connect(self._reset_setup)
        self.config_window.save_signal.connect(self._save_setting_changes)

        self.tray = SyncTrayIcon()
        self.tray.quit_signal.connect(self._stop)
        self.tray.login_signal.connect(self._show_login_window)
        self.tray.settings_signal.connect(self._show_config_window)
        self.tray.reset_setup_signal.connect(self._reset_setup)
        self.tray.sync_signal.connect(self._force_sync)
        self.tray.activated.connect(self._tray_icon_activated)
        self.tray.open_dir_signal.connect(self._open_download_dir)
        self.tray.show_menu_signal.connect(self._update_tray_menu)

        self.app.setQuitOnLastWindowClosed(False)

    def _setup_complete(self) -> None:
        self.setup_window.setVisible(False)
        self.model.setup(self.setup_window.institution_index,
                         self.setup_window.download_location,
                         self.setup_window.min_year)
        self._build_login_window(self.model.university.login)
        self._show_login_window()

    def _build_login_window(self, uni_login_info: InstitutionLogin) -> None:
        # Get login url from uni DB
        self.login_window = LoginWebView(start_url=str(uni_login_info.start_url),
                                         target_url=str(uni_login_info.target_url))
        self.login_window.login_complete_signal.connect(self._login_complete)

    def _login_complete(self) -> None:
        if self.login_window is None:
            return

        self.app.setOverrideCursor(Qt.WaitCursor)
        # Call login function on sync
        auth = self.model.auth(self.login_window.cookie_jar)
        self.tray.set_logged_in(auth)
        self.login_window.setVisible(False)
        self.app.restoreOverrideCursor()
        self._check_for_updates()
        self.tray.show_msg(*(self.tray_msg["download_started"]))

    def _reset_setup(self) -> None:
        # Hide login window and show setup wizard
        if self.login_window is not None:
            self._log_out()
            self.login_window.setVisible(False)
        self._show_setup_window()

    def _check_for_updates(self) -> None:
        if (html_url := check_for_updates()) is not None:
            if html_url == 'container':
                self.tray.show_msg(*(self.tray_msg["container_update"]))
            elif UpdateFoundDialog().should_update:
                webbrowser.open(html_url)

    def _show_login_window(self) -> None:
        if self.login_window is not None:
            self._show_window(self.login_window)

    def _show_setup_window(self) -> None:
        self._show_window(self.setup_window)

    def _show_config_window(self) -> None:
        # Update displayed settings
        self.config_window.download_location = self.model.download_location
        self.config_window.username = self.model.username.split(':')[1]
        self.config_window.sync_frequency = self.model.sync_interval
        self._show_window(self.config_window)

    def _show_window(self, window: QWidget) -> None:
        window.setWindowState(Qt.WindowNoState)
        window.show()
        window.setFocus()

    def _open_download_dir(self) -> None:
        # Open folder in browser
        OSUtils.open_dir_in_file_browser(self.model.download_location)

    def _tray_icon_activated(self, activation_reason: QSystemTrayIcon.ActivationReason) -> None:
        if activation_reason == QSystemTrayIcon.ActivationReason.Trigger:
            if self.model.university is None:
                self._show_setup_window()
            # if not logged in
            elif not self.model.is_logged_in:
                self._show_login_window()
        if self.model.has_error and not self._has_notified_error:
            self.tray.show_msg(*(self.tray_msg["download_error"]))
            webbrowser.open("https://github.com/jacobszpz/BlackboardSync/issues")
            self._has_notified_error = True

    def _log_out(self) -> None:
        if self.model.is_active:
            self.model.stop_sync()

        self.model.log_out()
        self.tray.set_logged_in(False)
        
        if self.login_window is not None:
            self.login_window.restore()
            self.login_window.setVisible(True)

        self.config_window.setVisible(False)

    def _save_setting_changes(self) -> None:
        self.config_window.setVisible(False)

        if self.model.download_location != self.config_window.download_location:
            redownload = RedownloadDialog().redownload
            self.model.change_download_location(self.config_window.download_location, redownload)

        self.model.sync_interval = self.config_window.sync_frequency

    def _force_sync(self) -> None:
        self.model.force_sync()

    def _update_tray_menu(self) -> None:
        # Update last sync time
        last_sync = self.model.last_sync_time
        last_sync_str = "Never"

        if last_sync is not None:
            last_sync_str = last_sync.strftime("%Y-%m-%d %H:%M:%S")

        self.tray.update_last_synced(last_sync_str)
        self.tray.set_logged_in(self.model.is_logged_in)
        if self.login_window is not None:
            self.login_window.setVisible(not self.model.is_logged_in)

        # Disable button if currently syncing
        self.tray.toggle_currently_syncing(self.model.is_syncing)

    def _stop(self) -> None:
        if self.model.is_active:
            self.model.stop_sync()
        self.app.quit()


if __name__ == '__main__':
    controller = BBSyncController()



"""Manages data of supported Blackboard partners."""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import logging
from typing import Optional
from pathlib import Path

import requests
from pydantic import HttpUrl, BaseModel

__all__ = ['InstitutionLogin', 'InstitutionNetwork', 'Institution',
           'get_by_index', 'get_index_by_ip', 'get_names']


# logger = logging.getLogger(__name__)

# Local institution data fallback in case of no server
_local_data = 'universities.json'
# Frequently updated remote datafile
# _remote_data = 'https://example.com/universities.json'
# IP API
_ip_api_endpoint = 'http://ip-api.com/json?fields=org,isp'


class InstitutionLogin(BaseModel):
    """Elements of login flow for an `Institution`.

    :param `HttpUrl` start_url: Link to the Blackboard portal.
    :param `HttpUrl` target_url: Landing URL after login is complete.
    :param str username_input_selector: Selector for username input field.
    :param str password_input_selector: Selector for password input field.
    """

    start_url: HttpUrl
    target_url: HttpUrl


class InstitutionNetwork(BaseModel):
    """Possible network values for an `Institution`.

    :param list[str] org: Possible values for the organization field.
    :param list[str] isp: Possible values for the ISP field.
    """

    org: list[str] = []
    isp: list[str] = []


class Institution(BaseModel):
    """Institution with a Blackboard portal.

    :param str name: Long-format name of institution.
    :param str short_name: Most common abbreviation of name.
    :param list[str] data_sources: List of Blackboard data sources to download by default.
    :param `InstitutionLogin` login: Login metadata.
    :param `InstitutionNetwork` network: Network metadata.
    """

    name: str
    short_name: Optional[str] = None
    data_sources: list[str] = []
    api_url: HttpUrl
    login: InstitutionLogin
    network: Optional[InstitutionNetwork] = None


def load() -> list[Institution]:
    remote_found = False
    json_data = {}

    # try:
    #   json_data = requests.get(_remote_data).json()
    # except (requests.ConnectionError, json.decoder.JSONDecodeError):
    #   remote_found = False

    if not remote_found:
        with (Path(__file__).parent / _local_data).open() as f:
            json_data = json.load(f)

    return [Institution(**i) for i in json_data]


_institutions = load()


def get_by_index(i: int) -> Institution:
    """Retrieve institution at index specified.

    :param int i: Index of institution.
    """
    return _institutions[i]


def get_names() -> list[str]:
    """Get institution names."""
    return [f"{ins.name} ({ins.short_name})" if ins.short_name is not None
            else ins.name for ins in _institutions]

def get_index_by_ip() -> Optional[int]:
    """Attempt to detect the user's institution by IP address data."""
    ip_info = {}

    try:
        ip_info = requests.get(_ip_api_endpoint).json()
    except (requests.ConnectionError, json.decoder.JSONDecodeError):
        pass

    ip_org = ip_info.get('org')
    ip_isp = ip_info.get('isp')

    for i, uni in enumerate(_institutions):
        if uni.network is None:
            continue

        found = False

        if ip_org:
            for org in uni.network.org:
                if ip_org == org:
                    found = True
        if ip_isp:
            for isp in uni.network.isp:
                if ip_isp == isp:
                    found = True
        if found:
            return i
    return None



"""
Blackboard Model Classes
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from enum import Enum
from datetime import datetime
from typing import Union, Optional

from pydantic import BaseModel, field_validator, ConfigDict
from pathvalidate import sanitize_filename


class ImmutableModel(BaseModel):
    """Model with const attributes."""
    model_config = ConfigDict(frozen=True)


class BBLocale(ImmutableModel):
    force: bool = False


class BBDurationType(str, Enum):
    """Blackboard Course Duration Type."""

    Continuous: str = 'Continuous'
    DateRange: str = 'DateRange'
    FixedNumDays: str = 'FixedNumDays'
    Term: str = 'Term'


class BBDuration(ImmutableModel):
    type: Optional[str] = None


class BBEnrollment(ImmutableModel):
    type: Optional[str] = None


class BBProctoring(ImmutableModel):
    secureBrowserRequiredToTake: bool = False
    secureBrowserRequiredToReview: bool = False
    webcamRequired: bool = False


class BBFile(ImmutableModel):
    """Blackboard File."""

    fileName: Optional[str] = None


class BBAttachment(ImmutableModel):
    """Blackboard File Attachment."""

    id: Optional[str] = None
    fileName: Optional[str] = None
    mimeType: Optional[str] = None


class BBLink(ImmutableModel):
    """Blackboard Link."""

    href: Optional[str] = None
    rel: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None


class BBAvailable(str, Enum):
    Yes = 'Yes'
    Term = 'Term'
    No = 'No'
    Disabled = 'Disabled'
    PartiallyVisible = 'PartiallyVisible'
    Other = '__blackboardsync_other'

    @classmethod
    def _missing_(cls, value: str) -> 'BBAvailable':
        return cls.Other

    def __bool__(self) -> bool:
        return not self in (BBAvailable.No, BBAvailable.Disabled)


class BBAvailability(ImmutableModel):
    available: Optional[BBAvailable] = None
    allowGuests: bool = False
    adaptiveRelease: dict = {}
    duration: Optional[BBDuration] = None


class BBMembership(ImmutableModel):
    """Blackboard Membership. Represents relation between student and course."""

    id: Optional[str] = None
    userId: Optional[str] = None
    courseId: Optional[str] = None
    dataSourceId: Optional[str] = None
    created: Optional[datetime] = None
    modified: Optional[datetime] = None
    availability: Optional[BBAvailability] = None
    courseRoleId: Optional[str] = None
    lastAccessed: Optional[datetime] = None
    childCourseId: Optional[str] = None


class BBResourceType(str, Enum):
    """Different resource types on Blackboard."""

    folder = 'x-bb-folder'
    file = 'x-bb-file'
    document = 'x-bb-document'
    externallink = 'x-bb-externallink'
    toollink = 'x-bb-toollink'
    turnitin_assignment = 'x-turnitin-assignment'
    bltiplacement_portal = 'x-bb-bltiplacement-Portal'
    assignment = 'x-bb-assignment'
    asmt_test_link = 'x-bb-asmt-test-link'
    syllabus = 'x-bb-syllabus'
    courselink = 'x-bb-courselink'
    blankpage = 'x-bb-blankpage'
    lesson = 'x-bb-lesson'
    other = '__blackboardsync_other'

    @classmethod
    def _missing_(cls, value):
        return cls.other


class BBContentHandler(ImmutableModel):
    id: Optional[BBResourceType] = None
    url: Optional[str] = None
    file: Optional[BBFile] = None
    gradeColumnId: Optional[str] = None
    groupContent: Optional[bool] = None
    targetId: Optional[str] = None
    targetType: Optional[str] = None
    placementHandle: Optional[str] = None
    assessmentId: Optional[str] = None
    proctoring: Optional[BBProctoring] = None

    @field_validator('id', mode='before')
    @classmethod
    def trim_resource_type(cls, v: str) -> str:
        return v.replace('resource/', '')

    @property
    def is_not_handled(self) -> bool:
        """Return true if resource should not be handled."""
        return self.id not in (BBResourceType.folder, BBResourceType.file,
                               BBResourceType.document, BBResourceType.externallink)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BBResourceType):
            return self.id == other
        elif isinstance(other, str):
            return self.id == BBResourceType(other)
        return False


class BBCourseContent(ImmutableModel):
    """Blackboard Content."""

    id: Optional[str] = None
    title: Optional[str] = None
    body: Optional[str] = None
    created: Optional[datetime] = None
    modified: Optional[datetime] = None
    position: int = 0
    hasChildren: bool = False
    launchInNewWindow: bool = False
    reviewable: bool = False
    availability: Optional[BBAvailability] = None
    contentHandler: Optional[BBContentHandler] = None
    links: list[BBLink] = []
    hasGradebookColumns: bool = False
    hasAssociatedGroups: bool = False

    def __str__(self):
        """Title of the course content."""
        return self.title

    @property
    def title_path_safe(self) -> str:
        """Return a path safe version of the title."""
        return sanitize_filename(self.title or 'Title missing',
                                 replacement_text='_') or ''


class BBContentChild(BBCourseContent):
    """Blackboard Content Child."""

    body: Optional[str] = None
    parentId: Optional[str] = None


class BBCourse(ImmutableModel):
    """BlackboardCourse. Represents an academic course."""

    _parse_name = True

    id: Optional[str] = None
    courseId: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    modified: Optional[datetime] = None
    organization: bool = False
    ultraStatus: Optional[str] = None
    closedComplete: bool = False
    availability: Optional[BBAvailability] = None
    enrollment: Optional[BBEnrollment] = None
    locale: Optional[BBLocale] = None
    externalAccessUrl: Optional[str] = None

    @property
    def code(self) -> Optional[str]:
        """Parse course code."""
        if self.name:
            code_split = self.name.split(' : ', 1)
            return code_split[0]
        return None

    @property
    def title(self) -> Optional[str]:
        """Parse course title."""
        if self.name:
            name_split = self.name.split(' : ', 1)[-1].split(',')
            return sanitize_filename(name_split[0], replacement_text='_')
        return None



"""
BlackboardSync.

Automatically sync content from Blackboard
"""

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import logging
import threading
from pathlib import Path
from typing import Optional
from datetime import datetime, timezone, timedelta

from requests.cookies import RequestsCookieJar
from requests.exceptions import ConnectionError
from pydantic import ValidationError

from .config import SyncConfig
from .download import BlackboardDownload
from .blackboard import BlackboardSession
from .institutions import Institution, get_by_index


class BlackboardSync:
    """Represents an instance of the BlackboardSync application."""

    _log_directory = "log"
    _app_name = 'blackboard_sync'

    # Seconds between each check of time elapsed since last sync
    _check_sleep_time = 10
    # Sync thread max retries
    _max_retries = 3

    _logger = logging.getLogger(__name__)

    def __init__(self):
        """Create an instance of the program."""

        # Download job
        self._download = None

        # Time between each sync in seconds
        self._sync_interval = 60 * 30

        # Time of next programmed sync
        self._next_sync = None
        # User session active
        self._is_logged_in = False

        # Flag to force sync
        self._force_sync = False
        # Flag to know if syncing is in progress
        self._is_syncing = False
        # Flag to know if syncing is on
        self._is_active = False
        # Flag to know if download thread has errors
        self._has_error = False

        # Set up logging
        self._logger.setLevel(logging.INFO)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.debug("Initialising BlackboardSync")

        sess_logger = logging.getLogger("BlackboardSession")
        sess_logger.setLevel(logging.WARN)
        sess_logger.addHandler(logging.StreamHandler())

        download_logger = logging.getLogger("BlackboardDownload")
        download_logger.setLevel(logging.DEBUG)
        download_logger.addHandler(logging.StreamHandler())

        self.sess_logger = sess_logger
        self.download_logger = download_logger

        self.university : Optional[Institution] = None
        self.sess : Optional[BlackboardSession] = None

        # Attempt to load existing configuration
        self._config = SyncConfig()
        self.logger.info("Loading preexisting configuration")

        if self._config.university_index is not None:
            self.university = get_by_index(self._config.university_index)

        if self._config.last_sync_time is not None:
            self._update_next_sync()

    def setup(self, university_index: int, download_location: Path,
              min_year: Optional[int] = None) -> None:
        """Setup the university information."""
        self.university_index = university_index
        self._config.download_location = download_location

        # If min_year has decreased, redownload
        if (self._config.min_year or 0) > (min_year or 0):
            self.last_sync_time = None

        self._config.min_year = min_year

    def auth(self, cookie_jar: RequestsCookieJar) -> bool:
        """Create a new Blackboard session with the given credentials.

        Will start syncing automatically if login successful.

        :param bool persistence: If true, login will be saved in the OS designated keyring.
        """
        if self.university is None:
            return False

        self._cookies = cookie_jar

        try:
            u_sess = BlackboardSession(str(self.university.api_url), cookie_jar)
        except ValueError:
            self.logger.warning("Credentials are incorrect")
        else:
            self.logger.info("Logged in successfully")
            self.sess = u_sess
            self._is_logged_in = True
            self.start_sync()

        return self._is_logged_in

    def log_out(self) -> None:
        """Stop syncing and forget user session."""
        self.stop_sync()
        self.sess = None
        self._is_logged_in = False

    def _sync_task(self) -> None:
        """Constantly check if the data is outdated and if so start a download job.

        Method run by Sync thread.
        """
        reload_session = False
        failed_attempts = 0

        while self._is_active:
            if self.outdated or self._force_sync:
                self.logger.debug("Syncing now")
                self._is_syncing = True

                # Download from last datetime
                user_session = self.sess

                if user_session is not None and self.university is not None:
                    self._download = BlackboardDownload(user_session,
                                                        self.download_location / '',
                                                        self.last_sync_time,
                                                        self.university.data_sources,
                                                        self.min_year)

                try:
                    if not self._is_active:
                        self._download.cancel()
                    job_start_time = self._download.download()
                    if job_start_time is not None:
                        self.last_sync_time = job_start_time
                    failed_attempts = 0
                except ValidationError as e:
                    self._log_exception(e)
                    self.logger.warning("Blackboard API validation failed")
                    self._has_error = True
                    self.stop_sync()
                except ValueError as e:
                    # Session (probably) expired, log out and attempt to reload config
                    self._log_exception(e)
                    self.logger.warning("User session expired")
                    reload_session = True
                    self.log_out()
                except ConnectionError as e:
                    # Random python connection error
                    self._log_exception(e)
                    failed_attempts += 1

                # Could not sync
                if failed_attempts >= self._max_retries:
                    self.log_out()

                # Reset force sync flag
                self._force_sync = False
                self._is_syncing = False
            if self._is_active:
                time.sleep(self._check_sleep_time)

        if reload_session:
            self.auth(self._cookies)

    def start_sync(self) -> bool:
        """Starts Sync thread or returns False if not possible."""
        if self._has_error:
            return False
        self.logger.info("Starting sync thread")
        self._is_active = True
        self.sync_thread = threading.Thread(target=self._sync_task)
        self.sync_thread.start()
        return True

    def stop_sync(self) -> None:
        """Stop Sync thread."""
        self.logger.info("Stopping sync thread")
        self._is_active = False

        if self._download is not None:
            self._download.cancel()

    def _log_exception(self, e: Exception) -> None:
        """Log exception to log file inside sync location."""
        self._log_dir.mkdir(exist_ok=True, parents=True)
        exception_log = logging.FileHandler(self._log_path)
        self.logger.addHandler(exception_log)
        self.logger.exception("Exception in sync thread")
        self.logger.removeHandler(exception_log)

    def force_sync(self) -> None:
        """Force Sync thread to start download job ASAP."""
        self.logger.debug("Forced syncing")
        self._force_sync = True

    @property
    def username(self) -> Optional[str]:
        return self.sess.username if self.sess is not None else None

    @property
    def _log_path(self) -> Path:
        filename = f"sync_error_{datetime.now():%Y-%m-%dT%H_%M_%S_%f}.log"
        return Path(self._log_dir / filename)

    @property
    def _log_dir(self) -> Path:
        return Path(self.download_location / self._log_directory)

    @property
    def last_sync_time(self) -> Optional[datetime]:
        """Datetime right before last download job started."""
        return self._config.last_sync_time

    @last_sync_time.setter
    def last_sync_time(self, last_time: Optional[datetime]):
        """Updates the last sync time recorded."""
        self._config.last_sync_time = last_time
        self._update_next_sync()

    def _update_next_sync(self) -> None:
        """Store a calculated datetime when next sync should take place."""
        if self.last_sync_time is not None:
            self._next_sync = (self._config.last_sync_time +
                               timedelta(seconds=self._sync_interval))

    @property
    def next_sync(self) -> datetime:
        """Time when last sync will be outdated."""
        return self._next_sync

    @property
    def outdated(self) -> bool:
        """Return true if last download job is outdated."""
        if self._config.last_sync_time is None:
            return True
        return datetime.now(timezone.utc) >= self.next_sync

    @property
    def min_year(self) -> int:
        return self._config.min_year

    @property
    def university_index(self):
        return self._config.university_index

    @university_index.setter
    def university_index(self, uni_index: int) -> None:
        self._config.university_index = uni_index
        self.university = get_by_index(uni_index)

    @property
    def download_location(self) -> Path:
        """Location to where all Blackboard content will be downloaded."""
        return self._config.download_location

    def change_download_location(self, new_dir: Path, redownload: bool = False) -> None:
        """Set new sync location.

        :param Path dir: The path of the sync dir.
        :param bool redownload: If true, ALL content will be re-downloaded to the new location.
        """
        if new_dir != self._config.download_location:
            # Update configuration
            self._config.download_location = new_dir

            if redownload:
                # Unset last sync time to fully download all files in new location
                self.last_sync_time = None

    @property
    def sync_interval(self) -> int:
        """Time to wait between download jobs."""
        return self._sync_interval

    @sync_interval.setter
    def sync_interval(self, p: int) -> None:
        self._sync_interval = p

    @property
    def is_active(self) -> bool:
        """Indicate the state of the sync thread."""
        return self._is_active

    @property
    def is_logged_in(self) -> bool:
        """Indicate if a user session is currently active."""
        return self._is_logged_in

    @property
    def is_syncing(self) -> bool:
        """Flag raised everytime a download job is running."""
        return self._is_syncing

    @property
    def has_error(self) -> bool:
        """Flag indicates an error resulting in no downloads."""
        return self._has_error

    @property
    def logger(self) -> logging.Logger:
        """Logger for BlackboardSync, set at level WARN."""
        return self._logger



from datetime import timezone

from pydantic import BaseModel
from hypothesis import given, infer
from hypothesis import strategies as st
from hypothesis.strategies import composite, DrawFn

from blackboard_sync.blackboard import (
    BBLocale, BBDuration, BBEnrollment, BBProctoring, BBFile, BBAttachment,
    BBCourseContent, BBAvailability)


handled_resource_types = ('x-bb-folder', 'x-bb-file', 'x-bb-document', 'x-bb-externallink')
unhandled_resource_types = ('x-bb-blankpage', 'x-bb-courselink', 'x-bb-syllabus', 'x-bb-toollink',
                            'x-bb-assignment', 'x-turnitin-assignment', 'x-bb-asmt-test-link',
                            'x-bb-bltiplacement-Portal', 'x-bb-bltiplacement-mediasite.lesopnames')
resource_types = handled_resource_types + unhandled_resource_types

class BlackboardCourseName(BaseModel):
    code: str
    title: str
    details: str

    @property
    def name(self):
        return f"{self.code} : {self.title}, {self.details}"


@composite
def bb_unhandled_resource_type(draw: DrawFn) -> str:
    res_type = draw(st.sampled_from(unhandled_resource_types))
    return f"resource/{res_type}"


@composite
def bb_handled_resource_type(draw: DrawFn) -> str:
    res_type = draw(st.sampled_from(handled_resource_types))
    return f"resource/{res_type}"


@composite
def bb_resource_type(draw: DrawFn) -> str:
    res_type = draw(st.sampled_from(resource_types))
    return f"resource/{res_type}"


@composite
def bb_course_code(draw: DrawFn) -> BlackboardCourseName:
    return draw(st.builds(BlackboardCourseName,
                          code=st.text().filter(lambda x: ':' not in x),
                          title=st.text().filter(lambda x: ',' not in x)))


st.register_type_strategy(BBLocale, st.builds(BBLocale, force=infer))
st.register_type_strategy(BBDuration, st.builds(BBDuration, type=infer))
st.register_type_strategy(BBEnrollment, st.builds(BBEnrollment, type=infer))
st.register_type_strategy(BBProctoring, st.builds(BBProctoring,
                                                  secureBrowserRequiredToTake=infer,
                                                  secureBrowserRequiredToReview=infer,
                                                  webcamRequired=infer))


@composite
def bb_proctorings(draw: DrawFn) -> BBProctoring:
    """Strategy to generate Blackboard Proctorings."""
    return draw()


@composite
def bb_file(draw: DrawFn) -> BBFile:
    """Strategy to generate Blackboard Files."""
    return draw(st.builds(BBFile, fileName=infer))


@composite
def bb_attachment(draw: DrawFn) -> BBAttachment:
    """Strategy to generate Blackboard Attachments"""
    return draw(st.builds(BBAttachment, id=infer, fileName=infer, mimeType=infer))


@composite
def bb_course_contents(draw: DrawFn) -> BBCourseContent:
    """A strategy to generate Blackboard Course Contents."""

    utc_strategy = st.just(timezone.utc)
    iso_datetime_strategy = st.datetimes(timezones=utc_strategy).map(lambda dt: dt.isoformat())

    draw(st.builds(BBCourseContent, id=infer, title=infer, body=infer,
                   created=iso_datetime_strategy, modified=iso_datetime_strategy,
                   position=infer, hasChildren=..., launchInNewWindow=infer,
                   reviewable=infer, availability=..., contentHandler=...,
                   links=..., hasGradebookColumns=infer, hasAssociatedGroups=infer))
    return 1


@composite
def bb_availabilities(draw: DrawFn) -> BBAvailability:
    """A strategy to generate Blackboard Availabities."""
    return draw(st.builds(BBAvailability, available="Yes"))



"""
BlackboardSync Qt GUI.

~~~~~~~~~~~~~~~~~~~~~

A graphical interface using Qt for BlackboardSync.

Basic usage:
    >>> from qt import LoginWebView
    >>> w = LoginWebView(start_url=..., target_url=...)


Full documentation available at <https://blackboardsync.readthedocs.io>.

:copyright: (c) 2021, Jacob Sánchez Pérez.
:license: GPL v2.0, see LICENSE for more details.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# Also available at https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

from .qt_elements import (SyncPeriod, SetupWizard, LoginWebView, SyncTrayIcon,
                          SyncTrayMenu, SettingsWindow, RedownloadDialog,
                          UniNotSupportedDialog, OSUtils, UpdateFoundDialog)

__all__ = ['SyncPeriod', 'SetupWizard', 'LoginWebView', 'SyncTrayIcon', 'SyncTrayMenu',
           'SettingsWindow', 'RedownloadDialog', 'UniNotSupportedDialog', 'OSUtils', 'UpdateFoundDialog']
